/***************************************************************************/
/*                                                                         */
/*  File: loadjpg.cpp                                                      */
/*  Author: bkenwright@xbdev.net                                           */
/*  Date: 19-01-06                                                         */
/*                                                                         */
/***************************************************************************/
/*
    About:
	Simplified jpg/jpeg decoder image loader - so we can take a .jpg file
	either from memory or file, and convert it either to a .bmp or directly
	to its rgb pixel data information.

	Simplified, and only deals with basic jpgs, but it covers all the
	information of how the jpg format works :)
*/
/***************************************************************************/


#include <stdio.h>		// sprintf(..), fopen(..)
#include <stdarg.h>     // So we can use ... (in dprintf)
#include <string.h>		// memset(..)


extern void dprintf(const char *fmt, ...);


/***************************************************************************/

#define DQT 	 0xDB	// Define Quantization Table
#define SOF 	 0xC0	// Start of Frame (size information)
#define DHT 	 0xC4	// Huffman Table
#define SOI 	 0xD8	// Start of Image
#define SOS 	 0xDA	// Start of Scan
#define EOI 	 0xD9	// End of Image, or End of File
#define APP0	 0xE0	

#define BYTE_TO_WORD(x) (((x)[0]<<8)|(x)[1])


#define HUFFMAN_TABLES		4
#define COMPONENTS			4
#define HUFFMAN_HASH_NBITS	9
#define HUFFMAN_HASH_SIZE	512

#define DCTSIZE	   8
#define DCTSIZE2   (DCTSIZE*DCTSIZE)

#define DEQUANTIZE(coef,quantval)  (((float) (coef)) * (quantval))

#define cY	1
#define cCb	2
#define cCr	3

/***************************************************************************/

static const unsigned char zigzag[64] = 
{
   0,  1,  5,  6, 14, 15, 27, 28,
   2,  4,  7, 13, 16, 26, 29, 42,
   3,  8, 12, 17, 25, 30, 41, 43,
   9, 11, 18, 24, 31, 40, 44, 53,
  10, 19, 23, 32, 39, 45, 52, 54,
  20, 22, 33, 38, 46, 51, 55, 60,
  21, 34, 37, 47, 50, 56, 59, 61,
  35, 36, 48, 49, 57, 58, 62, 63
};

/***************************************************************************/


struct huffman_table
{
  // Fast look up table, using HUFFMAN_HASH_NBITS bits we can have directly the symbol,
  // if the symbol is <0, then we need to look into the tree table
  short int lookup[HUFFMAN_HASH_SIZE];
  // code size: give the number of bits of a symbol is encoded
  unsigned char code_size[HUFFMAN_HASH_SIZE];
  // some place to store value that is not encoded in the lookup table 
  // FIXME: Calculate if 256 value is enough to store all values
  unsigned short slowtable[16-9][256];
};


struct component 
{
  unsigned int			m_Hfactor;
  unsigned int			m_Vfactor;
  float *				m_Q_table;			// Pointer to the quantisation table to use
  huffman_table*		AC_table;
  huffman_table*		DC_table;
  short int				m_previous_DC;		// Previous DC coefficient
  short int				m_DCT[64];			// DCT coef
};


struct jdec_data
{
	unsigned char*		m_rgb;				// Final Red Green Blue pixel data
	unsigned int		m_width;			// Width of image
	unsigned int		m_height;			// Height of image
	unsigned int		m_flags;

	const unsigned char*m_stream_begin;
	unsigned int		m_stream_length;

	const unsigned char*m_stream;			// Pointer to the current stream
	unsigned int reservoir;
	unsigned int nbits_in_reservoir;

	component			m_component_infos[COMPONENTS];
	float				m_Q_tables[COMPONENTS][64];	// quantization tables
	huffman_table		m_HTDC[HUFFMAN_TABLES];		// DC huffman tables  
	huffman_table		m_HTAC[HUFFMAN_TABLES];		// AC huffman tables

	// Temp space used after the IDCT to store each components
	unsigned char		m_Y[64*4];
	unsigned char		m_Cr[64];
	unsigned char		m_Cb[64];

	// Internal Pointer use for colorspace conversion, do not modify it !!!
	unsigned char *		m_colourspace;
};


/***************************************************************************/
// 
//  Returns the size of the file in bytes
//
/***************************************************************************/
inline int FileSize(FILE *fp)
{
	long pos;
	fseek(fp, 0, SEEK_END);
	pos = ftell(fp);
	fseek(fp, 0, SEEK_SET);
	return pos;
}

/***************************************************************************/

// Clamp our integer between 0 and 255
inline unsigned char Clamp(int i)
{
	if (i<0)
		return 0;
	else if (i>255)
		return 255;
	else
		return i;
}   

/***************************************************************************/

inline unsigned char DeScaleAndClamp(float xx, int shift)
{
	int x = (int)xx;

	x += (1UL<<(shift-1));
	if (x<0)
		x = (x >> shift) | ((~(0UL)) << (32-(shift)));
	else
		x >>= shift;

	x += 128;
	x = Clamp(x);	// Clamp 0 to 255
	return x;
}


/***************************************************************************/
//
// Perform dequantization and inverse DCT on one block of coefficients.
//
/***************************************************************************/
inline void IDCT (component *compptr, unsigned char *output_buf, int stride)
{
	float tmp0, tmp1, tmp2, tmp3, tmp4, tmp5, tmp6, tmp7;
	float tmp10, tmp11, tmp12, tmp13;
	float z5, z10, z11, z12, z13;
	short *inptr;
	float *quantptr;
	float *wsptr;
	unsigned char *outptr;
	int ctr;
	float workspace[DCTSIZE2]; // buffers data between passes

	// Pass 1: process columns from input, store into work array.

	inptr = compptr->m_DCT;
	quantptr = compptr->m_Q_table;
	wsptr = workspace;
	for (ctr = DCTSIZE; ctr > 0; ctr--) 
	{
		// Due to quantization, we will usually find that many of the input
		// coefficients are zero, especially the AC terms.  We can exploit this
		// by short-circuiting the IDCT calculation for any column in which all
		// the AC terms are zero.  In that case each output is equal to the
		// DC coefficient (with scale factor as needed).
		// With typical images and quantization tables, half or more of the
		// column DCT calculations can be simplified this way.
     
    
		if (inptr[DCTSIZE*1] == 0 && inptr[DCTSIZE*2] == 0 &&
		inptr[DCTSIZE*3] == 0 && inptr[DCTSIZE*4] == 0 &&
		inptr[DCTSIZE*5] == 0 && inptr[DCTSIZE*6] == 0 &&
		inptr[DCTSIZE*7] == 0) 
		{
			// AC terms all zero
			float dcval = DEQUANTIZE(inptr[DCTSIZE*0], quantptr[DCTSIZE*0]);

			wsptr[DCTSIZE*0] = dcval;
			wsptr[DCTSIZE*1] = dcval;
			wsptr[DCTSIZE*2] = dcval;
			wsptr[DCTSIZE*3] = dcval;
			wsptr[DCTSIZE*4] = dcval;
			wsptr[DCTSIZE*5] = dcval;
			wsptr[DCTSIZE*6] = dcval;
			wsptr[DCTSIZE*7] = dcval;
      
			inptr++;			// advance pointers to next column
			quantptr++;
			wsptr++;
			continue;
		}
    
		// Even part

		tmp0 = DEQUANTIZE(inptr[DCTSIZE*0], quantptr[DCTSIZE*0]);
		tmp1 = DEQUANTIZE(inptr[DCTSIZE*2], quantptr[DCTSIZE*2]);
		tmp2 = DEQUANTIZE(inptr[DCTSIZE*4], quantptr[DCTSIZE*4]);
		tmp3 = DEQUANTIZE(inptr[DCTSIZE*6], quantptr[DCTSIZE*6]);

		tmp10 = tmp0 + tmp2;	// phase 3
		tmp11 = tmp0 - tmp2;

		tmp13 = tmp1 + tmp3;	// phases 5-3
		tmp12 = (tmp1 - tmp3) * ((float) 1.414213562) - tmp13; /* 2*c4 */

		tmp0 = tmp10 + tmp13;	// phase 2
		tmp3 = tmp10 - tmp13;
		tmp1 = tmp11 + tmp12;
		tmp2 = tmp11 - tmp12;
    
		// Odd part

		tmp4 = DEQUANTIZE(inptr[DCTSIZE*1], quantptr[DCTSIZE*1]);
		tmp5 = DEQUANTIZE(inptr[DCTSIZE*3], quantptr[DCTSIZE*3]);
		tmp6 = DEQUANTIZE(inptr[DCTSIZE*5], quantptr[DCTSIZE*5]);
		tmp7 = DEQUANTIZE(inptr[DCTSIZE*7], quantptr[DCTSIZE*7]);

		z13 = tmp6 + tmp5;		// phase 6
		z10 = tmp6 - tmp5;
		z11 = tmp4 + tmp7;
		z12 = tmp4 - tmp7;

		tmp7 = z11 + z13;		// phase 5
		tmp11 = (z11 - z13) * ((float) 1.414213562);	// 2*c4

		z5 = (z10 + z12) * ((float) 1.847759065);		// 2*c2
		tmp10 = ((float) 1.082392200) * z12 - z5;		// 2*(c2-c6)
		tmp12 = ((float) -2.613125930) * z10 + z5;		// -2*(c2+c6)

		tmp6 = tmp12 - tmp7;	// phase 2
		tmp5 = tmp11 - tmp6;
		tmp4 = tmp10 + tmp5;

		wsptr[DCTSIZE*0] = tmp0 + tmp7;
		wsptr[DCTSIZE*7] = tmp0 - tmp7;
		wsptr[DCTSIZE*1] = tmp1 + tmp6;
		wsptr[DCTSIZE*6] = tmp1 - tmp6;
		wsptr[DCTSIZE*2] = tmp2 + tmp5;
		wsptr[DCTSIZE*5] = tmp2 - tmp5;
		wsptr[DCTSIZE*4] = tmp3 + tmp4;
		wsptr[DCTSIZE*3] = tmp3 - tmp4;

		inptr++;			// advance pointers to next column
		quantptr++;
		wsptr++;
	}
  
	// Pass 2: process rows from work array, store into output array.
	// Note that we must descale the results by a factor of 8 == 2**3.

	wsptr = workspace;
	outptr = output_buf;
	for (ctr = 0; ctr < DCTSIZE; ctr++) 
	{
		// Rows of zeroes can be exploited in the same way as we did with columns.
		// However, the column calculation has created many nonzero AC terms, so
		// the simplification applies less often (typically 5% to 10% of the time).
		// And testing floats for zero is relatively expensive, so we don't bother.

    
		// Even part

		tmp10 = wsptr[0] + wsptr[4];
		tmp11 = wsptr[0] - wsptr[4];

		tmp13 = wsptr[2] + wsptr[6];
		tmp12 = (wsptr[2] - wsptr[6]) * ((float) 1.414213562) - tmp13;

		tmp0 = tmp10 + tmp13;
		tmp3 = tmp10 - tmp13;
		tmp1 = tmp11 + tmp12;
		tmp2 = tmp11 - tmp12;

		// Odd part

		z13 = wsptr[5] + wsptr[3];
		z10 = wsptr[5] - wsptr[3];
		z11 = wsptr[1] + wsptr[7];
		z12 = wsptr[1] - wsptr[7];

		tmp7 = z11 + z13;
		tmp11 = (z11 - z13) * ((float) 1.414213562);

		z5 = (z10 + z12) * ((float) 1.847759065);  // 2*c2 
		tmp10 = ((float) 1.082392200) * z12 - z5;  // 2*(c2-c6) 
		tmp12 = ((float) -2.613125930) * z10 + z5; // -2*(c2+c6)

		tmp6 = tmp12 - tmp7;
		tmp5 = tmp11 - tmp6;
		tmp4 = tmp10 + tmp5;

		// Final output stage: scale down by a factor of 8 and range-limit

		outptr[0] = DeScaleAndClamp(tmp0 + tmp7, 3);
		outptr[7] = DeScaleAndClamp(tmp0 - tmp7, 3);
		outptr[1] = DeScaleAndClamp(tmp1 + tmp6, 3);
		outptr[6] = DeScaleAndClamp(tmp1 - tmp6, 3);
		outptr[2] = DeScaleAndClamp(tmp2 + tmp5, 3);
		outptr[5] = DeScaleAndClamp(tmp2 - tmp5, 3);
		outptr[4] = DeScaleAndClamp(tmp3 + tmp4, 3);
		outptr[3] = DeScaleAndClamp(tmp3 - tmp4, 3);


		wsptr += DCTSIZE;		// advance pointer to next row
		outptr += stride;
	}
}


/***************************************************************************/

/***************************************************************************/
//
// Save a buffer in 24bits Bitmap (.bmp) format 
//
/***************************************************************************/
inline void WriteBMP24(const char* szBmpFileName, int Width, int Height, unsigned char* RGB)
{
	#pragma pack(1)
	struct stBMFH // BitmapFileHeader & BitmapInfoHeader
	{
		// BitmapFileHeader
		char         bmtype[2];     // 2 bytes - 'B' 'M'
		unsigned int iFileSize;     // 4 bytes
		short int    reserved1;     // 2 bytes
		short int    reserved2;     // 2 bytes
		unsigned int iOffsetBits;   // 4 bytes
		// End of stBMFH structure - size of 14 bytes
		// BitmapInfoHeader
		unsigned int iSizeHeader;    // 4 bytes - 40
		unsigned int iWidth;         // 4 bytes
		unsigned int iHeight;        // 4 bytes
		short int    iPlanes;        // 2 bytes
		short int    iBitCount;      // 2 bytes
		unsigned int Compression;    // 4 bytes
		unsigned int iSizeImage;     // 4 bytes
		unsigned int iXPelsPerMeter; // 4 bytes
		unsigned int iYPelsPerMeter; // 4 bytes
		unsigned int iClrUsed;       // 4 bytes
		unsigned int iClrImportant;  // 4 bytes
		// End of stBMIF structure - size 40 bytes
		// Total size - 54 bytes
	};
	#pragma pack()

	// Round up the width to the nearest DWORD boundary
	int iNumPaddedBytes = 4 - (Width * 3) % 4;
	iNumPaddedBytes = iNumPaddedBytes % 4;

	stBMFH bh;
	memset(&bh, 0, sizeof(bh));
	bh.bmtype[0]='B';
	bh.bmtype[1]='M';
	bh.iFileSize = (Width*Height*3) + (Height*iNumPaddedBytes) + sizeof(bh);
	bh.iOffsetBits = sizeof(stBMFH);
	bh.iSizeHeader = 40;
	bh.iPlanes = 1;
	bh.iWidth = Width;
	bh.iHeight = Height;
	bh.iBitCount = 24;


	char temp[1024]={0};
	sprintf(temp, "%s", szBmpFileName);
	FILE* fp = fopen(temp, "wb");
	fwrite(&bh, sizeof(bh), 1, fp);
	for (int y=Height-1; y>=0; y--)
	{
		for (int x=0; x<Width; x++)
		{
			int i = (x + (Width)*y) * 3;
			unsigned int rgbpix = (RGB[i]<<16)|(RGB[i+1]<<8)|(RGB[i+2]<<0);
			fwrite(&rgbpix, 3, 1, fp);
		}
		if (iNumPaddedBytes>0)
		{
			unsigned char pad = 0;
			fwrite(&pad, iNumPaddedBytes, 1, fp);
		}
	}
	fclose(fp);
}

/***************************************************************************/
// Takes two array of bits, and build the huffman table for size, and code
//
// lookup will return the symbol if the code is less or equal than HUFFMAN_HASH_NBITS.
// code_size will be used to known how many bits this symbol is encoded.
// slowtable will be used when the first lookup didn't give the result.
/***************************************************************************/
inline void BuildHuffmanTable(const unsigned char *bits, const unsigned char *vals, huffman_table *table)
{
	unsigned int i, j, code, code_size, val, nbits;
	unsigned char huffsize[257], *hz;
	unsigned int huffcode[257], *hc;
	int next_free_entry;

	// Build a temp array 
	// huffsize[X] => numbers of bits to write vals[X]
	hz = huffsize;
	for (i=1; i<=16; i++)
	{
		for (j=1; j<=bits[i]; j++)
		{
			*hz++ = i;
		}
	}
	*hz = 0;

	memset(table->lookup, 0xff, sizeof(table->lookup));
	for (i=0; i<(16-HUFFMAN_HASH_NBITS); i++)
	{
		table->slowtable[i][0] = 0;
	}

	// Build a temp array
	// huffcode[X] => code used to write vals[X]
	code = 0;
	hc = huffcode;
	hz = huffsize;
	nbits = *hz;
	while (*hz)
	{
		while (*hz == nbits) 
		{
			*hc++ = code++;
			hz++;
		}
		code <<= 1;
		nbits++;
	}

	// Build the lookup table, and the slowtable if needed.	
	next_free_entry = -1;
	for (i=0; huffsize[i]; i++)
	{
		val = vals[i];
		code = huffcode[i];
		code_size = huffsize[i];

		dprintf("val=%2.2x code=%8.8x codesize=%2.2d\n", i, code, code_size);

		table->code_size[val] = code_size;
		if (code_size <= HUFFMAN_HASH_NBITS)
		{
			// Good: val can be put in the lookup table, so fill all value of this
			// column with value val 
			int repeat = 1UL<<(HUFFMAN_HASH_NBITS - code_size);
			code <<= HUFFMAN_HASH_NBITS - code_size;
			while ( repeat-- )
			{
				table->lookup[code++] = val;
			}

		}
		else
		{
			// Perhaps sorting the array will be an optimization
			unsigned short *slowtable = table->slowtable[code_size-HUFFMAN_HASH_NBITS-1];
			while(slowtable[0])
			{
				slowtable+=2;
			}
			slowtable[0] = code;
			slowtable[1] = val;
			slowtable[2] = 0;
			// TODO: NEED TO CHECK FOR AN OVERFLOW OF THE TABLE
		}

   }
}

/***************************************************************************/

inline void PrintSOF(const unsigned char *stream)
{
	int width;
	int height;
	int nr_components;
	int precision;

	const char *nr_components_to_string[] = 
										{   "????",
											"Grayscale",
											"????",
											"YCbCr",
											"CYMK" };

	precision = stream[2];
	height = BYTE_TO_WORD(stream+3);
	width  = BYTE_TO_WORD(stream+5);
	nr_components = stream[7];

	dprintf("> SOF marker\n");
	dprintf("Size:%dx%d nr_components:%d (%s)  precision:%d\n", 
							width, height,
							nr_components, 
							nr_components_to_string[nr_components],
							precision);
}

/***************************************************************************/

inline int ParseSOF(jdec_data *jdata, const unsigned char *stream)
{
	int i, width, height, nr_components, cid, sampling_factor;
	int Q_table;
	component *c;

	PrintSOF(stream);

	height = BYTE_TO_WORD(stream+3);
	width  = BYTE_TO_WORD(stream+5);
	nr_components = stream[7];

	stream += 8;
	for (i=0; i<nr_components; i++) 
	{
		cid = *stream++;
		sampling_factor = *stream++;
		Q_table = *stream++;
		c = &jdata->m_component_infos[cid];
		c->m_Vfactor = sampling_factor&0xf;
		c->m_Hfactor = sampling_factor>>4;
		c->m_Q_table = jdata->m_Q_tables[Q_table];

		dprintf("Component:%d  factor:%dx%d  Quantization table:%d\n",
				cid, 
				c->m_Hfactor, 
				c->m_Hfactor, 
				Q_table );
	}
	jdata->m_width = width;
	jdata->m_height = height;

	return 0;
}

/***************************************************************************/

inline void BuildQuantizationTable(float *qtable, const unsigned char *ref_table)
{
	// For float AA&N IDCT method, divisors are equal to quantization
	// coefficients scaled by scalefactor[row]*scalefactor[col], where
	// scalefactor[0] = 1
	// scalefactor[k] = cos(k*PI/16) * sqrt(2)    for k=1..7
	// We apply a further scale factor of 8.
	// What's actually stored is 1/divisor so that the inner loop can
	// use a multiplication rather than a division.
	int i, j;
	static const double aanscalefactor[8] = 
	{
		1.0, 1.387039845, 1.306562965, 1.175875602,
		1.0, 0.785694958, 0.541196100, 0.275899379
	};
	const unsigned char *zz = zigzag;

	for (i=0; i<8; i++) 
	{
		for (j=0; j<8; j++) 
		{
			*qtable++ = (float)(ref_table[*zz++] * aanscalefactor[i] * aanscalefactor[j]);
		}
	}
}

/***************************************************************************/

inline int ParseDQT(jdec_data *jdata, const unsigned char *stream)
{
	int length, qi;
	float *table;

	dprintf("> DQT marker\n");
	length = BYTE_TO_WORD(stream) - 2;
	stream += 2;	// Skip length

	while (length>0)
	{
		qi = *stream++;

		if (qi>>4)
		{
			dprintf("16 bits quantization table is not supported\n");
		}
		if (qi>4)
		{
			dprintf("No more 4 quantization table is supported (got %d)\n", qi);
		}

		table = jdata->m_Q_tables[qi];
		BuildQuantizationTable(table, stream);
		stream += 64;
		length -= 65;
	}
	return 0;
}

/***************************************************************************/

inline int ParseSOS(jdec_data *jdata, const unsigned char *stream)
{
	unsigned int i, cid, table;
	unsigned int nr_components = stream[2];

	dprintf("> SOS marker\n");

	if (nr_components != 3)
	{
		dprintf("We only support YCbCr image\n");
	}


	stream += 3;
	for (i=0;i<nr_components;i++) 
	{
		cid = *stream++;
		table = *stream++;

		if ((table&0xf)>=4)
		{
			dprintf("We do not support more than 2 AC Huffman table\n");
		}
		if ((table>>4)>=4)
		{
			dprintf("We do not support more than 2 DC Huffman table\n");
		}
		dprintf("ComponentId:%d  tableAC:%d tableDC:%d\n", cid, table&0xf, table>>4);

		jdata->m_component_infos[cid].AC_table = &jdata->m_HTAC[table&0xf];
		jdata->m_component_infos[cid].DC_table = &jdata->m_HTDC[table>>4];
	}
	jdata->m_stream = stream+3;
	return 0;
}

/***************************************************************************/

inline int ParseDHT(jdec_data *jdata, const unsigned char *stream)
{
	unsigned int count, i;
	unsigned char huff_bits[17];
	int length, index;

	length = BYTE_TO_WORD(stream) - 2;
	stream += 2;	// Skip length

	dprintf("> DHT marker (length=%d)\n", length);

	while (length>0) 
	{
		index = *stream++;

		// We need to calculate the number of bytes 'vals' will takes
		huff_bits[0] = 0;
		count = 0;
		for (i=1; i<17; i++) 
		{
			huff_bits[i] = *stream++;
			count += huff_bits[i];
		}

		if (count > 1024)
		{
			dprintf("No more than 1024 bytes is allowed to describe a huffman table");
		}
		if ( (index &0xf) >= HUFFMAN_TABLES)
		{
			dprintf("No mode than %d Huffman tables is supported\n", HUFFMAN_TABLES);
		}
		dprintf("Huffman table %s n%d\n", (index&0xf0)?"AC":"DC", index&0xf);
		dprintf("Length of the table: %d\n", count);


		if (index & 0xf0 )
		{
			BuildHuffmanTable(huff_bits, stream, &jdata->m_HTAC[index&0xf]);
		}
		else
		{
			BuildHuffmanTable(huff_bits, stream, &jdata->m_HTDC[index&0xf]);
		}

		length -= 1;
		length -= 16;
		length -= count;
	}
	dprintf("< DHT marker\n");
	return 0;
}


inline int ParseJFIF(jdec_data *jdata, const unsigned char *stream)
{
	int chuck_len;
	int marker;
	int sos_marker_found = 0;
	int dht_marker_found = 0;

	// Parse marker
	while (!sos_marker_found)
	{
		if (*stream++ != 0xff)
		{
			goto bogus_jpeg_format;
		}

		// Skip any padding ff byte (this is normal)
		while (*stream == 0xff)
		{
			stream++;
		}

		marker = *stream++;
		chuck_len = BYTE_TO_WORD(stream);

		switch (marker)
		{
			case SOF:
			{
				if (ParseSOF(jdata, stream) < 0)
					return -1;
			}
			break;

			case DQT:
			{
				if (ParseDQT(jdata, stream) < 0)
					return -1;
			}
			break;

			case SOS:
			{
				if (ParseSOS(jdata, stream) < 0)
					return -1;
				sos_marker_found = 1;
			}
			break;

			case DHT:
			{
				if (ParseDHT(jdata, stream) < 0)
					return -1;
				dht_marker_found = 1;
			}
			break;

			// The reason I added these additiona skips here, is because for
			// certain jpg compressions, like swf, it splits the encoding 
			// and image data with SOI & EOI extra tags, so we need to skip
			// over them here and decode the whole image
			case SOI:
			case EOI:
			{
				chuck_len = 0;
				break;
			}
			break;

			case APP0:
			{
				dprintf("APP0 Chunk (txt information) skipping\n");
			}
			break;

			default:
			{
				dprintf("ERROR> Unknown marker %2.2x\n", marker);
			}
			break;
		}

		stream += chuck_len;
	}

	if (!dht_marker_found) 
	{
		dprintf("ERROR> No Huffman table loaded\n");
	}

	return 0;

	bogus_jpeg_format:
	dprintf("ERROR> Bogus jpeg format\n");
	return -1;
}

/***************************************************************************/

inline int JpegParseHeader(jdec_data *jdata, const unsigned char *buf, unsigned int size)
{
	int ret;

	// Identify the file
	if ((buf[0] != 0xFF) || (buf[1] != SOI))
	{
		dprintf("Not a JPG file ?\n");
		return -1;
	}

	jdata->m_stream_begin = buf+2;
	jdata->m_stream_length = size-2;

	ret = ParseJFIF(jdata, jdata->m_stream_begin);

	return ret;
}

/***************************************************************************/

inline void JpegGetImageSize(jdec_data *jdata, unsigned int *width, unsigned int *height)
{
	*width = jdata->m_width;
	*height = jdata->m_height;
}


/***************************************************************************/

inline void fill_nbits(unsigned int& reservoir,unsigned int& nbits_in_reservoir, const unsigned char** stream, int& nbits_wanted)
{
	while ((int)nbits_in_reservoir<nbits_wanted)
	{
		const unsigned char c = *(*stream)++;
		reservoir <<= 8;
		if (c == 0xff && (**stream) == 0x00)
			(*stream)++;
		reservoir |= c;
		nbits_in_reservoir+=8;
	}
}

inline void get_nbits(unsigned int& reservoir, unsigned int& nbits_in_reservoir, const unsigned char** stream, int nbits_wanted, short& result)
{
	fill_nbits(reservoir,nbits_in_reservoir,stream,(nbits_wanted));
	result = ((reservoir)>>(nbits_in_reservoir-(nbits_wanted))); 
	nbits_in_reservoir -= (nbits_wanted); 
	reservoir &= ((1U<<nbits_in_reservoir)-1);
	if (result < (short)(1UL<<((nbits_wanted)-1)))
	{
		result = result + (short)(0xFFFFFFFFUL<<(nbits_wanted))+1;
	}
}

inline void look_nbits(unsigned int& reservoir, unsigned int& nbits_in_reservoir, const unsigned char** stream, int nbits_wanted, int& result)
{
	fill_nbits(reservoir,nbits_in_reservoir,stream,(nbits_wanted));
	result = ((reservoir)>>(nbits_in_reservoir-(nbits_wanted)));
}

inline void skip_nbits(unsigned int& reservoir, unsigned int& nbits_in_reservoir, const unsigned char** stream, int& nbits_wanted)
{
	fill_nbits(reservoir,nbits_in_reservoir,stream,(nbits_wanted));
	nbits_in_reservoir -= (nbits_wanted); 
	reservoir &= ((1U<<nbits_in_reservoir)-1);
}

/***************************************************************************/

inline int GetNextHuffmanCode(jdec_data *jdata, huffman_table *huffman_table)
{
	int value, hcode;
	int extra_nbits, nbits;
	unsigned short *slowtable;

	look_nbits(jdata->reservoir, jdata->nbits_in_reservoir, &jdata->m_stream, HUFFMAN_HASH_NBITS, hcode);
	value = huffman_table->lookup[hcode];
	if (value>=0)
	{ 
		int code_size = huffman_table->code_size[value];
		skip_nbits(jdata->reservoir, jdata->nbits_in_reservoir, &jdata->m_stream, code_size);
		return value;
	}

	/* Decode more bits each time ... */
	for (extra_nbits=0; extra_nbits<16-HUFFMAN_HASH_NBITS; extra_nbits++)
	{
		nbits = HUFFMAN_HASH_NBITS + 1 + extra_nbits;

		look_nbits(jdata->reservoir, jdata->nbits_in_reservoir, &jdata->m_stream, nbits, hcode);
		slowtable = huffman_table->slowtable[extra_nbits];
		/* Search if the code is in this array */
		while (slowtable[0]) 
		{
			if (slowtable[0] == hcode) 
			{
				skip_nbits(jdata->reservoir, jdata->nbits_in_reservoir, &jdata->m_stream, nbits);
				return slowtable[1];
			}
			slowtable+=2;
		}
	}
	return 0;
}

/***************************************************************************/

inline void ProcessHuffmanDataUnit(jdec_data *jdata, int componentA)
{
	unsigned char j;
	int huff_code;
	unsigned char size_val, count_0;

	component *c = &jdata->m_component_infos[componentA];
	short int DCT[64];

	// Initialize the DCT coef table
	memset(DCT, 0, sizeof(DCT));

	// DC coefficient decoding
	huff_code = GetNextHuffmanCode(jdata, c->DC_table);
	if (huff_code) 
	{
		get_nbits(jdata->reservoir, jdata->nbits_in_reservoir, &jdata->m_stream, huff_code, DCT[0]);
		DCT[0] += c->m_previous_DC;
		c->m_previous_DC = DCT[0];
	} 
	else 
	{
		DCT[0] = c->m_previous_DC;
	}

	// AC coefficient decoding

	j = 1;
	while (j<64)
	{
		huff_code = GetNextHuffmanCode(jdata, c->AC_table);

		size_val = huff_code & 0xF;
		count_0 = huff_code >> 4;

		if (size_val == 0)
		{// RLE 
			if (count_0 == 0)
				break;	// EOB found, go out
			else if (count_0 == 0xF)
			j += 16;	// skip 16 zeros
		}
		else
		{
			j += count_0;	// skip count_0 zeroes
			get_nbits(jdata->reservoir, jdata->nbits_in_reservoir, &jdata->m_stream, size_val, DCT[j]);
			j++;
		}
	}

	for (j = 0; j < 64; j++)
	{
		c->m_DCT[j] = DCT[zigzag[j]];
	}
}


/***************************************************************************/

inline void ConvertYCrCbtoRGB(int y, int cb, int cr,
							  int* r, int* g, int* b)

{
	float red, green, blue;

	red   = y + 1.402f*(cb-128);
	green = y-0.34414f*(cr-128)-0.71414f*(cb-128);
	blue  = y+1.772f*(cr-128);

	*r = (int) Clamp((int)red);
	*g = (int) Clamp((int)green);
	*b = (int) Clamp((int)blue);
}

/***************************************************************************/

inline void YCrCB_to_RGB24(jdec_data *jdata, int w, int h)
{
	const unsigned char *Y, *Cb, *Cr;
	unsigned char *pix;

	int r, g, b;

	Y  = jdata->m_Y;
	Cb = jdata->m_Cb;
	Cr = jdata->m_Cr;

//	dprintf("***pix***\n\n");
	for (int y=0; y<8*h; y++)
	{
		for (int x=0; x<8*w; x++)
		{
			int poff = x*3 + jdata->m_width*3*y;
			pix = &(jdata->m_colourspace[poff]);
			
			int yoff = x + y*(w*8);
			int coff = (int)(x*(1.0f/w)) + (int)(y*(1.0f/h))*8;

			int yc =  Y[yoff];
			int cb = Cb[coff];
			int cr = Cr[coff];

			ConvertYCrCbtoRGB(yc,cr,cb,&r,&g,&b);

			pix[0] = Clamp(r);
			pix[1] = Clamp(g);
			pix[2] = Clamp(b);

//			dprintf("-[%d][%d][%d]-\t", poff, yoff, coff);
		}
//		dprintf("\n");
	}
//	dprintf("\n\n");
}

/***************************************************************************/
//
// Decoding
//  .-------.
//  | 1 | 2 |
//  |---+---|
//  | 3 | 4 |
//  `-------'
//
/***************************************************************************/
inline void decode_MCU_planes(jdec_data *jdata, int w, int h)
{
	// Y
	for (int y=0; y<h; y++)
	{
		for (int x=0; x<w; x++)
		{
			int stride = w*8;
			int offset = x*8 + y*64*w;

			ProcessHuffmanDataUnit(jdata, cY);
			IDCT(&jdata->m_component_infos[cY], &jdata->m_Y[offset], stride);
		}
	}

	// Cb
	ProcessHuffmanDataUnit(jdata, cCb);
	IDCT(&jdata->m_component_infos[cCb], jdata->m_Cb, 8);

	// Cr
	ProcessHuffmanDataUnit(jdata, cCr);
	IDCT(&jdata->m_component_infos[cCr], jdata->m_Cr, 8);
}


/***************************************************************************/

inline void ReSync(jdec_data *jdata)
{
	// Init DC coefficients
	for (int i=0; i<COMPONENTS; i++)
	{
		jdata->m_component_infos[i].m_previous_DC = 0;
	}

	jdata->reservoir = 0;
	jdata->nbits_in_reservoir = 0;
}

/***************************************************************************/

inline int JpegDecode(jdec_data *jdata)
{
	int hFactor = jdata->m_component_infos[cY].m_Hfactor;
	int vFactor = jdata->m_component_infos[cY].m_Vfactor;

	// RGB24:
	if (jdata->m_rgb == NULL)
	{
		int h = jdata->m_height*3;
		int w = jdata->m_width*3;
		int height = h + (8*hFactor) - (h%(8*hFactor));
		int width  = w + (8*vFactor) - (w%(8*vFactor));
		jdata->m_rgb = new unsigned char[width * height];
	}

	ReSync(jdata);

	int xstride_by_mcu = 8*hFactor;
	int ystride_by_mcu = 8*vFactor;

	// Don't forget to that block can be either 8 or 16 lines
	unsigned int bytes_per_blocklines = jdata->m_width*3 * ystride_by_mcu;

	unsigned int bytes_per_mcu = 3*xstride_by_mcu;

	// Just the decode the image by macroblock (size is 8x8, 8x16, or 16x16)
	int yy = 0;
	for (int y=0; y < (int)jdata->m_height; y+=ystride_by_mcu)
	{
		//dprintf("Decoding row %d\n", y);
		jdata->m_colourspace = jdata->m_rgb + (yy * bytes_per_blocklines);

		for (int x=0; x < (int)(jdata->m_width); x+=xstride_by_mcu)
		{
			decode_MCU_planes(jdata, hFactor, vFactor );

			YCrCB_to_RGB24(jdata, hFactor, vFactor );

			jdata->m_colourspace += bytes_per_mcu;
		}
		yy++;
	}

	return 0;
}

/***************************************************************************/
//
// Take Jpg data, i.e. jpg file read into memory, and decompress it to an
// array of rgb pixel values.
//
/***************************************************************************/
int DecodeJpgFileData(const unsigned char* buf, int sizeBuf, unsigned char** rgbpix, unsigned int* width, unsigned int* height)
{
	// Allocate memory for our decoded jpg structure, all our data will be
	// decompressed and stored in here for the various stages of our jpeg decoding
	jdec_data* jdec = new jdec_data();
	if (jdec == NULL)
	{
		dprintf("Not enough memory to alloc the structure need for decompressing\n");
		return 0;
	}

	// Start Parsing.....reading & storing data
	if (JpegParseHeader(jdec, buf, sizeBuf)<0)
	{
		dprintf("ERROR > parsing jpg header\n");
	}

	// We've read it all in, now start using it, to decompress and create rgb values
	dprintf("Decoding JPEG image...\n");
	JpegDecode(jdec);

	// Get the size of the image
	JpegGetImageSize(jdec, width, height);

	*rgbpix = jdec->m_rgb;

	// Release the memory for our jpeg decoder structure jdec
	delete jdec;

	return 1;
}

/***************************************************************************/
//
// Load one jpeg image, and decompress it, and save the result.
//
/***************************************************************************/
int ConvertJpgFile(char* szJpgFileInName, char * szBmpFileOutName)
{
	FILE *fp;
	unsigned int lengthOfFile;
	unsigned char *buf;
	

	// Load the Jpeg into memory
	fp = fopen(szJpgFileInName, "rb");
	if (fp == NULL)
	{
		dprintf("Cannot open jpg file: %s\n", szJpgFileInName);
		return 0;
	}

	lengthOfFile = FileSize(fp);
	buf = new unsigned char[lengthOfFile + 4];
	if (buf == NULL)
	{
		dprintf("Not enough memory for loading file\n");
		return 0;
	}
	fread(buf, lengthOfFile, 1, fp);
	fclose(fp);

	unsigned char* rgbpix = NULL;
	unsigned int width  = 0;
	unsigned int height = 0;
	DecodeJpgFileData(buf, lengthOfFile, &rgbpix, &width, &height);

	if (rgbpix==NULL)
	{
		dprintf("Failed to decode jpg\n");
		return 0;
	}

	// Delete our data we read in from the file
	delete[] buf;

	// Save it
	WriteBMP24(szBmpFileOutName, width, height, rgbpix);

	// Since we don't need the pixel information anymore, we must
	// release this as well
	delete[] rgbpix;

	return 1;
}

/***************************************************************************/

